"""This module contains functions to automatically add tapers to a component's ports, and to create tapers between different cross sections."""

from __future__ import annotations

import warnings

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.port import Port
from gdsfactory.typings import CrossSectionSpec, LayerTransitions, Ports


def add_auto_tapers(
    component: Component,
    ports: Ports,
    cross_section: CrossSectionSpec,
) -> list[Port]:
    """Adds tapers to the ports of a component (to be used for routing) and returns the new lists of ports.

    Args:
        component: the component to add tapers to
        ports: the list of ports
        cross_section: the cross section to route to

    Returns:
        The new list of ports, on the opposite end of the tapers
    """
    cross_section_obj = gf.get_cross_section(cross_section)
    cs_layer = gf.get_layer(cross_section_obj.layer)
    cs_width = cross_section_obj.width

    layer_transitions = None
    _pdk = None

    def get_layer_transitions() -> LayerTransitions:
        nonlocal layer_transitions, _pdk
        if layer_transitions is None:
            if _pdk is None:
                _pdk = gf.get_active_pdk()
            layer_transitions = _pdk.layer_transitions
        return layer_transitions

    result = []
    get_port_layer = gf.get_layer

    for p in ports:
        port_layer = get_port_layer(p.layer)
        port_width = p.width
        lt = layer_transitions
        if lt is None:
            lt = get_layer_transitions()
        reverse = False

        if port_layer != cs_layer:
            try:
                taper = lt.get((port_layer, cs_layer))
                if taper is None:
                    taper = lt[cs_layer, port_layer]
                    reverse = True
            except KeyError as e:
                raise KeyError(
                    f"No registered tapers between routing layers {gf.get_layer_name(port_layer)!r} and {gf.get_layer_name(cs_layer)!r}!"
                ) from e
        elif port_width != cs_width:
            try:
                taper = lt[port_layer]
            except KeyError:
                warnings.warn(
                    f"No registered width taper for layer {port_layer}. Skipping.",
                    stacklevel=4,
                )
                result.append(p)
                continue
        else:
            result.append(p)
            continue

        if reverse:
            taper_component = gf.get_component(
                taper, width2=port_width, width1=cs_width
            )
        else:
            taper_component = gf.get_component(
                taper, width1=port_width, width2=cs_width
            )

        # Ensure we filter out i.e. extra electrical ports if we should be looking at optical ones
        taper_ports = [
            tp for tp in taper_component.ports if tp.port_type == p.port_type
        ]

        if len(taper_ports) != 2:
            raise ValueError(
                f"Taper component should have two ports of port_type={p.port_type!r}! Got {taper_component.ports}."
            )
        if taper_ports[0].layer == p.layer and taper_ports[0].width == p.width:
            p0, p1 = taper_ports
        elif taper_ports[1].layer == p.layer and taper_ports[1].width == p.width:
            p1, p0 = taper_ports
        else:
            width = p.width
            layer = p.layer
            raise ValueError(
                f"Taper component ports do not match the port's layer and width!\nTaper name: {taper_component.name}\nPorts: {taper_ports}\nCross-section: {layer=}, {width=}"
            )
        taper_ref = component.add_ref(taper_component)
        assert p0.name is not None
        taper_ref.connect(p0.name, p)
        port_new = taper_ref.ports[p1.name].copy()
        port_new.name = p.name
        result.append(port_new)
    return result


def auto_taper_to_cross_section(
    component: gf.Component,
    port: Port,
    cross_section: CrossSectionSpec,
    layer_transitions: LayerTransitions | None = None,
) -> Port:
    """Creates a taper from a port to a given cross section and places it in the component. The opposite port of the taper will be returned.

    Args:
        component: the component to place into
        port: a port to connect to, usually from a ComponentReference
        cross_section: a cross section to transition to
        layer_transitions: the layer transitions dictionary to use (use the pdk default if None)

    Returns:
        The port at the opposite (unconnected end) of the taper.
    """
    # Reuse fast version above for legacy compatibility
    # (not used in new add_auto_tapers flow)
    port_layer = gf.get_layer(port.layer)
    port_width = port.width
    cross_section_obj = gf.get_cross_section(cross_section)
    cs_layer = gf.get_layer(cross_section_obj.layer)
    cs_width = cross_section_obj.width
    if layer_transitions is None:
        pdk = gf.get_active_pdk()
        layer_transitions = pdk.layer_transitions
    reverse = False
    if port_layer != cs_layer:
        try:
            taper = layer_transitions.get((port_layer, cs_layer))
            if taper is None:
                taper = layer_transitions[cs_layer, port_layer]
                reverse = True
        except KeyError as e:
            raise KeyError(
                f"No registered tapers between routing layers {gf.get_layer_name(port_layer)!r} and {gf.get_layer_name(cs_layer)!r}!"
            ) from e
    elif port_width != cs_width:
        try:
            taper = layer_transitions[port_layer]
        except KeyError:
            warnings.warn(
                f"No registered width taper for layer {port_layer}. Skipping.",
                stacklevel=4,
            )
            return port
    else:
        return port
    if reverse:
        taper_component = gf.get_component(taper, width2=port_width, width1=cs_width)
    else:
        taper_component = gf.get_component(taper, width1=port_width, width2=cs_width)

    # ensure we filter out i.e. extra electrical ports if we should be looking at optical ones
    taper_ports = [p for p in taper_component.ports if p.port_type == port.port_type]

    if len(taper_ports) != 2:
        raise ValueError(
            f"Taper component should have two ports of port_type={port.port_type!r}! Got {taper_component.ports}."
        )
    if taper_ports[0].layer == port.layer and taper_ports[0].width == port.width:
        p0, p1 = taper_ports
    elif taper_ports[1].layer == port.layer and taper_ports[1].width == port.width:
        p1, p0 = taper_ports
    else:
        width = port.width
        layer = port.layer
        raise ValueError(
            f"Taper component ports do not match the port's layer and width!\nTaper name: {taper_component.name}\nPorts: {taper_ports}\nCross-section: {layer=}, {width=}"
        )
    taper_ref = component.add_ref(taper_component)
    assert p0.name is not None
    taper_ref.connect(p0.name, port)
    port_new = taper_ref.ports[p1.name].copy()
    port_new.name = port.name
    return port_new



"""Install Klayout and GIT plugins."""

from __future__ import annotations

import configparser
import os
import pathlib
import shutil
import sys

from git import Repo

from gdsfactory.config import PATH

home = pathlib.Path.home()


def clone_repository(repo_url: str, clone_dir: pathlib.Path) -> None:
    try:
        Repo.clone_from(repo_url, clone_dir)
        print(f"Repository cloned to {clone_dir}")
    except Exception as e:
        print("Error cloning repository:", e)


def remove_path_or_dir(dest: pathlib.Path) -> None:
    if dest.is_symlink():
        os.unlink(dest)
    elif dest.is_dir():
        shutil.rmtree(dest)
    else:
        os.remove(dest)


def make_link(src: pathlib.Path, dest: pathlib.Path, overwrite: bool = True) -> None:
    dest = pathlib.Path(dest)
    if dest.exists() and not overwrite:
        print(f"{dest} already exists")
        return
    if dest.exists() or dest.is_symlink():
        print(f"removing {dest} already installed")
        remove_path_or_dir(dest)
    try:
        os.symlink(src, dest, target_is_directory=True)
        print("Symlink made:")
    except OSError as err:
        print("Could not create symlink!")
        print("     Error: ", err)
        if sys.platform == "win32":
            # https://stackoverflow.com/questions/32877260/privlege-error-trying-to-create-symlink-using-python-on-windows-10
            shutil.copytree(src, dest)
            print("Copied directory:")
    print(f"From: {src}")
    print(f"To:   {dest}")


def install_gdsdiff() -> None:
    """Install gdsdiff tool for GIT."""
    print("gdsdiff shows boolean differences in Klayout so you can run:")
    print("git diff FILE.GDS")
    print("Adding gdsdiff command to ~/.gitconfig and ~/.config/git/attributes")
    _write_git_config()
    _write_git_attributes()


def _write_git_config() -> None:
    """Write GIT config in ~/.gitconfig."""
    git_config_path = home / ".gitconfig"
    config = configparser.RawConfigParser()
    config.read(git_config_path)
    key = 'diff "gdsdiff"'

    if key not in config.sections():
        config.add_section(key)
        config.set(key, "command", "python -m gdsfactory.difftest_git")
        config.set(key, "binary", "True")

        with open(git_config_path, "w+") as f:
            config.write(f, space_around_delimiters=True)


def _write_git_attributes() -> None:
    """Write git attributes in ~/.config/git/attributes."""
    git_config = home / ".config" / "git"
    git_config.mkdir(exist_ok=True, parents=True)
    line_to_add = "*.gds diff=gdsdiff\n"

    # Specify the path to the .config/git/attributes file
    dirpath = home / ".config" / "git"
    dirpath.mkdir(exist_ok=True, parents=True)
    file_path = dirpath / "attributes"

    # Read the file to check if the line already exists
    file_content = file_path.read_text() if file_path.exists() else ""

    # Add the line only if it doesn't exist
    if line_to_add not in file_content:
        with open(file_path, "a") as file:
            file.write(line_to_add)


def get_klayout_path() -> pathlib.Path:
    """Returns KLayout path."""
    klayout_folder = "KLayout" if sys.platform == "win32" else ".klayout"
    return home / klayout_folder


def copy(src: pathlib.Path, dest: pathlib.Path) -> None:
    """Copy overwriting file or directory."""
    dest_folder = dest.parent
    dest_folder.mkdir(exist_ok=True, parents=True)

    if dest.exists() or dest.is_symlink():
        print(f"removing {dest} already installed")
        remove_path_or_dir(dest)

    if src.is_dir():
        shutil.copytree(src, dest)
    else:
        shutil.copy(src, dest)
    print(f"{src} copied to {dest}")


def _install_to_klayout(
    src: pathlib.Path, klayout_subdir_name: str, package_name: str
) -> None:
    """Install into KLayout technology.

    Equivalent to using KLayout package manager.

    """
    klayout_folder = "KLayout" if sys.platform == "win32" else ".klayout"
    subdir = home / klayout_folder / klayout_subdir_name
    dest = subdir / package_name
    subdir.mkdir(exist_ok=True, parents=True)
    make_link(src, dest)


def install_klayout_package() -> None:
    """Install gdsfactory KLayout package.

    Equivalent to using KLayout package manager.
    """
    cwd = pathlib.Path(__file__).resolve().parent

    # install layermap
    _install_to_klayout(
        src=cwd / "generic_tech" / "klayout",
        klayout_subdir_name="salt",
        package_name="gdsfactory",
    )

    klayout_folder = "KLayout" if sys.platform == "win32" else ".klayout"
    subdir = home / klayout_folder / "salt"

    # install metainfo-ports
    clone_repository(
        "git@github.com:gdsfactory/metainfo-ports.git", subdir / "metainfo-ports"
    )

    # install klive
    clone_repository("git@github.com:gdsfactory/klive.git", subdir / "klive")


def install_klayout_technology(
    tech_dir: pathlib.Path, tech_name: str | None = None
) -> None:
    """Install technology to KLayout."""
    _install_to_klayout(
        src=tech_dir,
        klayout_subdir_name="tech",
        package_name=tech_name or tech_dir.name,
    )


py_files = list(PATH.notebooks.glob("**/*.py"))


def convert_py_to_ipynb(
    files: list[pathlib.Path] = py_files,
    output_folder: pathlib.Path = PATH.cwd / "notebooks",
) -> None:
    """Convert notebooks from markdown to ipynb."""
    import jupytext

    output_folder.mkdir(exist_ok=True, parents=True)

    for file in files:
        notebook_file = f"{output_folder}/{file.stem}.ipynb"
        nb = jupytext.read(file)
        jupytext.write(nb, notebook_file)


if __name__ == "__main__":
    # cwd = pathlib.Path(__file__).resolve().parent
    # home = pathlib.Path.home()
    # src = cwd / "generic_tech" / "klayout" / "tech"

    # write_git_attributes()
    # install_gdsdiff()
    install_klayout_package()
    # convert_py_to_ipynb()
